/*
 * d_submatrix.c
 *
 * $Source: /source/MASTER/garden/datlib/RCS/d_submatrix.c,v $
 * $Date: 91/11/02 22:34:44 $
 * $Author: klee $
 * $Locker:  $
 *
 * This file is part of the dat library.
 *
 * $Log:	d_submatrix.c,v $
 * Revision 1.2  91/11/02  22:34:44  klee
 * found lying around...
 * 
 * Revision 1.1  90/12/08  03:57:10  kaboom
 * Initial revision
 * 
 */

#include "dat.h"
#include "datio.h"
#include <sys/types.h>
#ifndef LINT
static char rcsid[]="$Header: /source/MASTER/garden/datlib/RCS/d_submatrix.c,v 1.2 91/11/02 22:34:44 klee Exp $";
#endif

#define STRICT 1		/* strict handling of out-of-range args */

#ifdef APPEND_MODE
/* start = -1 means a special call from d_read, d_write, or d_tell */
#endif

long 
  d_submatrix(df, start, stop, step)
DATFILE *df;
struct d_addr *start, *stop, *step;
{
  int fd, d, xd;
  long dim[2+MAXD];
  off_t sizeproduct, lseek();
  char *malloc();
  register struct dinfo *ip;
#ifdef APPEND_MODE
  long elem;
  off_t where=0;
#endif
  
  fd = fileno(d_getfp(df));
  if ((ip=_Dinfo[fd]) == NULLI) 
    {
      ip = _Dinfo[fd] = (struct dinfo *) malloc(sizeof (struct dinfo));
      if (ip == NULLI)
	return(_daterrh("d_submatrix: out of core for %s\n",
			d_getname(df)));
    }
	
  ip->ndims = d_getdim(df, dim, dim+1, dim+2, MAXD) + 2;
  if (ip->ndims<3)
    return(_daterrh("d_submatrix: dimensions of %s unknown\n",
		    d_getname(df)));
  
  ip->dtype = d_gettype(df, (int *) NULL);
  ip->lbps = _dsize(ip->dtype);
  ip->eof = 0;
  
  sizeproduct = 1 << ip->lbps;
  ip->offset = 0L;
  
#ifdef APPEND_MODE
  if (start == (struct d_addr *)-1) {			/* record current place */
    where = lseek(fd, 0L, 1);
    if (where % (1<<ip->lbps) != 0)
      return(_daterrh("d_submatrix: misaligned i/o pointer on %s\n",
		      d_getname(df)));
    start = NULLA;
  }
  elem = where >> ip->lbps;
#endif
  
  /* process dimensions from fastest varying index first;
     need xd since tell[], subdim[], and skip[] are backwards
     from datfile standard */
  
  for (d=0, xd=ip->ndims-1; xd>=0; d++, xd--) {
    long size, start1, stop1, step1;
    start1 = start ? ((long *)start)[xd] : 0;
    stop1 = stop ? ((long *)stop)[xd] : dim[xd];
    step1 = step ? ((long *)step)[xd] : 1;
    
#ifdef STRICT								/* stop if out-of-range args */
    if (step1<=0 || start1<0 || start1>=stop1 || stop1>dim[xd])
      return(_daterrh("d_submatrix: arguments out of range for %s\n",
		      d_getname(df)));
#else										/* fix up out-of-range args */
    if (step1 <= 0) 
      step1 = 1;

    if (start1 < 0) 
      start1 = 0;
    else if (start1 >= dim[xd]) 
      start1 = dim[xd]-1;

    if (stop1 > dim[xd]) 
      stop1 = dim[xd];
    else if (stop1 <= start1) 
      stop1 = start1+1;
#endif
    ip->basis[d] = step1 * sizeproduct;
    ip->subdim[d] = size = (stop1 - start1 + step1 - 1) / step1;
    /* BUG: When the following "optimization" is triggered by size==1 && step1>1,
       one lseek will be inadvertently bypassed in _dread and _dwrite. */
    /*		if (d == 0) ip->hsub = (size > 1) ? step1 : 1; */
    if (d == 0) ip->hsub = step1;
    else ip->nskip[d-1] = ip->basis[d] - ip->basis[d-1]*ip->subdim[d-1];
    
#ifdef APPEND_MODE
    ip->tell[d] = elem % size;
    elem /= size;
#else
    ip->tell[d] = 0;
#endif
    ip->offset += start1 * sizeproduct;
    sizeproduct *= dim[xd];
  }
  
#ifdef APPEND_MODE
  if (elem != 0) ip->eof = 1;
  /* seek to start of subarray if where!=0 */
  return((where ? where : lseek(fd, ip->offset, 0)) >> ip->lbps);
#else
  /* seek to start of subarray */
  return(lseek(fd, ip->offset, 0) >> ip->lbps);
#endif
}
